# frozen_string_literal: true

require_relative './action/flash'
require_relative './action/parameters'
require_relative './action/routes'

module Swamp
  module Action
    def self.included(base)
      base.extend ClassMethods
      base.include Swamp::Action::Parameters
      base.include Swamp::Action::Routes
      base.include Swamp::Action::Flash
    end

    class Result
      attr_reader :template, :redirection_path, :redirection_options

      def render(template)
        @template = template
      end

      def redirect_to(path, **options)
        @redirection_path = path
        @redirection_options = options
      end

      def prepare!(payload)
        exposures.merge!(payload)
        self
      end

      def redirect?
        !redirection_path.nil?
      end

      def render?
        !template.nil?
      end

      def exposures
        @exposures ||= Hash[]
      end

      protected

      def method_missing(method_name, *)
        exposures.fetch(method_name) { super }
      end

      def respond_to_missing?(method_name, _include_all)
        exposures.key?(method_name.to_sym) ? true : super
      end
    end

    def call(args = Hash[])
      prepare(args)

      catch :done do
        _call
      end

      finalize(args)
    end

    def _call; end

    private

    def prepare(kwargs)
      @request = kwargs[:request]
      @__result = ::Swamp::Action::Result.new
    end

    def expose(*names)
      names.each do |name|
        send(name) if self.class.exposures.include?(name)
      end
    end

    def finalize(*)
      @__result.prepare!(_exposures)
    end

    def _exposures
      Hash[].tap do |result|
        self.class.exposures.each do |name, ivar|
          result[name] = respond_to?(name) ? send(name) : instance_variable_get(ivar)
        end
      end
    end

    def render(template)
      @__result.render(template)
      throw :done
    end

    def redirect_to(path, **options)
      @__result.redirect_to(path, options)
      throw :done
    end

    module ClassMethods
      attr_reader :exposures

      def self.extended(interactor)
        interactor.class_eval do
          self.exposures = {}
        end
      end

      def call(args = Hash[])
        new.call(args)
      end

      def expose(*instance_variable_names)
        instance_variable_names.each do |name|
          exposures[name.to_sym] = "@#{name}"
        end
      end

      protected

      attr_writer :exposures
    end
  end
end
